/*
 * ligkern.c
 *
 * This file is part of the ttf2pk package.
 *
 *    Frederic Loyer <loyer@ensta.fr>
 *    Werner Lemberg <a7971428@unet.univie.ac.at>
 */

#include <string.h>

#include "ttf2tfm.h"
#include "ligkern.h"
#include "ttfenc.h"
#include "texenc.h"
#include "newobj.h"
#include "errormsg.h"


static char *
paramstring(char **curp)
{
  register char *p, *q;


  p = *curp;
  while (*p > ' ')
    p++;
  q = *curp;
  if (*p != '\0')
    *p++ = '\0';
  while (*p && *p <= ' ')
    p++;
  *curp = p;
  return q;
}


/*
 *   Some routines to remove kerns that match certain patterns.
 */

static struct kern *
rmkernmatch(struct kern *k,
            char *s)
{
  struct kern *nk;


  while (k && strcmp(k->succ, s) == 0)
    k = k->next;

  if (k)
  {
    for (nk = k; nk; nk = nk->next)
      while (nk->next && strcmp(nk->next->succ, s) == 0)
        nk->next = nk->next->next;
  }
  return k;
}


/*
 *   Recursive to one level.
 */

static void
rmkern(char *s1, char *s2,
       struct ttfinfo *ai,
       struct Font *fnt)
{
  if (ai == NULL)
  {
    if (strcmp(s1, "*") == 0)
    {
      for (ai = fnt->charlist; ai; ai = ai->next)
        rmkern(s1, s2, ai, fnt);
      return;
    }
    else
    {
      ai = findadobe(s1, fnt->charlist);
      if (ai == NULL)
        return;
    }
  }

  if (strcmp(s2, "*") == 0)
    ai->kerns = NULL; /* drop them on the floor */
  else
    ai->kerns = rmkernmatch(ai->kerns, s2);
}


/*
 *   Make the kerning for character S1 equivalent to that for S2.
 *   If either S1 or S2 do not exist, do nothing.
 *   If S1 already has kerning, do nothing.
 */

static void
addkern(char *s1, char *s2,
        struct Font *fnt)
{
  struct ttfinfo *ai1 = findadobe(s1, fnt->charlist);
  struct ttfinfo *ai2 = findadobe(s2, fnt->charlist);


  if (ai1 && ai2 && !ai1->kerns)
  {
    /* Put the new one at the head of the list, since order is immaterial. */
    struct ttfptr *ap =
      (struct ttfptr *)mymalloc((unsigned long)sizeof (struct ttfptr));


    ap->next = ai2->kern_equivs;
    ap->ch = ai1;
    ai2->kern_equivs = ap;
  }
}


/*
 *   Reads a ligkern line, if this is one.  Assumes the first character
 *   passed is `%'.
 */

void
checkligkern(char *s, struct Font *fnt, int *sawligkernp, char **pos)
{
  char *opos = *pos;
  char *mlist[5];
  int n;


  s++;
  while (*s && *s <= ' ')
    s++;
  if (strncmp(s, "LIGKERN", 7) == 0)
  {
    *sawligkernp = 1;
    s += 7;
    while (*s && *s <= ' ')
      s++;
    *pos = s;
    while (**pos)
    {
      for (n = 0; n < 5;)
      {
        if (**pos == '\0')
          break;
        mlist[n] = paramstring(pos);
        if (strcmp(mlist[n], ";") == 0)
          break;
        n++;
      }

      if (n > 4)
        oops("Too many parameters in lig kern data");
      if (n < 3)
        oops("Too few parameters in lig kern data");

      if (n == 3 && strcmp(mlist[1], "{}") == 0)        /* rmkern command */
        rmkern(mlist[0], mlist[2], (struct ttfinfo *)0, fnt);
      else if (n == 3 && strcmp(mlist[1], "<>") == 0)   /* addkern */
        addkern(mlist[0], mlist[2], fnt);
      else if (n == 3 && strcmp(mlist[0], "||") == 0 &&
               strcmp(mlist[1], "=") == 0)              /* bc command */
      {
        struct ttfinfo *ai = findadobe("||", fnt->charlist);


        if (fnt->boundarychar != -1)
          oops("Multiple boundary character commands?");
        if (sscanf(mlist[2], "%d", &n) != 1)
          oops("Expected number assignment for boundary char");
        if (n < 0 || n > 0xFF)
          oops("Boundary character number must be 0..0xFF");

        fnt->boundarychar = n;
        if (ai == NULL)
          oops("Internal error: boundary char");
        ai->outcode = n; /* prime the pump, so to speak, for lig/kerns */
      }
      else if (n == 4)
      {
        int op = -1;
        struct ttfinfo *ai;


        for (n = 0; encligops[n]; n++)
          if (strcmp(mlist[2], encligops[n]) == 0)
          {
            op = n;
            break;
          }
        if (op < 0)
          oops("Bad ligature op specified");

        if (NULL != (ai = findadobe(mlist[0], fnt->charlist)))
        {
          struct lig *lig;


          if (findadobe(mlist[2], fnt->charlist))
                                                /* remove coincident kerns */
            rmkern(mlist[0], mlist[1], ai, fnt);

          if (strcmp(mlist[3], "||") == 0)
            oops("You can't lig to the boundary character!");

          if (!fnt->fixedpitch)         /* fixed pitch fonts get *0* ligs */
          {
            for (lig = ai->ligs; lig; lig = lig->next)
              if (strcmp(lig->succ, mlist[1]) == 0)
                break;                  /* we'll re-use this structure */

            if (lig == NULL)
            {
              lig = newlig();
              lig->succ = newstring(mlist[1]);
              lig->next = ai->ligs;
              ai->ligs = lig;
            }
            lig->sub = newstring(mlist[3]);
            lig->op = op;
 
            if (strcmp(mlist[1], "||") == 0)
            {
              lig->boundleft = 1;
              if (strcmp(mlist[0], "||") == 0)
                oops("You can't lig boundarychar boundarychar!");
            }
            else
              lig->boundleft = 0;
          }
        }
      }
      else
        oops("Bad form in LIGKERN command");
    }
  }
  *pos = opos;
}


void
getligkerndefaults(struct Font *fnt,
                   int *sawligkernp,
                   char *buffer, char *obuffer,
                   char **curp)
{
  int i;


  for (i = 0; staticligkern[i]; i++)
  {
    strcpy(buffer, staticligkern[i]);
    strcpy(obuffer, staticligkern[i]);
    *curp = buffer;
    checkligkern(buffer, fnt, sawligkernp, curp);
  }
}


/* end */
